@@ -1,16 +1,20 @@
 
 from generalfile import Path
 from generalpackager import GIT_PASSWORD
+from generallibrary import Ver
 
 from setuptools import find_namespace_packages
 import re
 from git import Repo
+import subprocess
+import sys
 
 
 class LocalRepo:
     """ Tools to help Path interface a Local Python Repository. """
+    enabled = ...
     name = ...
-    version = ...
+    version = ...  # type: Ver
     description = ...
     install_requires = ...
     extras_require = ...
@@ -24,14 +28,11 @@ class LocalRepo:
         self.path = Path(path).absolute()
         self.git_exclude_lines = git_exclude_lines
 
-        self.name = self.path.parts()[-1]
+        metadata = {"enabled": True}
+        metadata.update(self.get_metadata_path().read())
 
-        for key, value in self.get_metadata_path().read().items():
+        for key, value in metadata.items():
             setattr(self, f"_{key}", value)
-            setattr(LocalRepo, key, property(
-                lambda self, key=key: getattr(self, f"_{key}", ...),
-                lambda self, value, key=key: LocalRepo.metadata_setter(self, value, key),
-            ))
 
         for key in self.metadata_keys:
             if getattr(self, key) is Ellipsis:
@@ -41,11 +42,31 @@ class LocalRepo:
             self.extras_require["full"] = list(set().union(*self.extras_require.values()))
             self.extras_require["full"].sort()
 
-    def metadata_setter(self, value, key):
+        self.version = Ver(self.version)
+
+    @staticmethod
+    def get_repos_path(path):
+        """ Try to return repos path by iterating parents if None. """
+        if path is None:
+            repos_path = Path.get_working_dir()
+            while not LocalRepo.get_local_repos(repos_path):
+                repos_path = repos_path.get_parent()
+                if repos_path is None:
+                    raise AttributeError(f"Couldn't find repos path.")
+            return repos_path
+        else:
+            return Path(path).absolute()
+
+    @classmethod
+    def is_creatable(cls, path):
+        """ Return whether this API can be created. """
+        return cls.path_is_repo(path=path)
+
+    def metadata_setter(self, key, value):
         """ Set a metadata's key both in instance and json file. """
         if value != getattr(self, f"_{key}"):
             metadata = self.get_metadata_path().read()
-            metadata[key] = value
+            metadata[key] = str(value)
             self.get_metadata_path().write(metadata, overwrite=True, indent=4)
 
         setattr(self, f"_{key}", value)
@@ -74,9 +95,9 @@ class LocalRepo:
         """ Get a Path instance pointing to workflow.yml, regardless if it exists. """
         return self.path / ".github/workflows/workflow.yml"
 
-    def get_test_main_path(self):
+    def get_test_path(self):
         """ Get a Path instance pointing to workflow.yml, regardless if it exists. """
-        return self.path / f"{self.name}/test/main.py"
+        return self.path / f"{self.name}/test"
 
     def get_package_paths(self):
         """ Get a list of Paths pointing to each folder containing a Python file in this local repo, aka `namespace package`. """
@@ -85,13 +106,16 @@ class LocalRepo:
     @classmethod
     def get_local_repos(cls, folder_path):
         """ Return a list of local repos in given folder. """
-        return [path for path in Path(folder_path).get_paths_in_folder() if cls.path_is_repo(path)]
+        folder_path = Path(folder_path)
+        if not folder_path.exists():
+            return []
+        return [path for path in folder_path.get_paths_in_folder() if cls.path_is_repo(path)]
 
     @classmethod
     def path_is_repo(cls, path):
         """ Return whether this path is a local repo. """
         path = Path(path)
-        if path.is_file():
+        if path.is_file() or not path.exists(quick=True):
             return False
         for file in path.get_paths_in_folder():
             if file.name() in ("metadata.json", "setup.py"):
@@ -104,9 +128,8 @@ class LocalRepo:
             :rtype: dict[list[str]] """
         todos = []
         for path in self.path.get_paths_recursive():
-            if path.name().lower() in ("shelved.patch", "readme.md") or any([exclude in path for exclude in self.git_exclude_lines]):
+            if path.name().lower() in ("shelved.patch", "readme.md") or any([exclude.replace("*", "") in path for exclude in self.git_exclude_lines]):
                 continue
-
             try:
                 text = path.text.read()
             except:
@@ -119,25 +142,61 @@ class LocalRepo:
                 })
         return todos
 
-    def commit_and_push(self, message=None):
-        """ Commit and push this local repo to GitHub. """
+    def commit_and_push(self, message=None, tag=False):
+        """ Commit and push this local repo to GitHub.
+            Return short sha1 of pushed commit. """
         if message is None:
             message = "Automatic commit."
 
         repo = Repo(str(self.path))
+
         repo.git.add(A=True)
-        repo.index.commit(message=message)
+        repo.index.commit(message=str(message))
         remote = repo.remote()
         remote.set_url(f"https://Mandera:{GIT_PASSWORD}@github.com/ManderaGeneral/{self.name}.git")
-        remote.push()
+
+        if tag:
+            tag_ref = repo.create_tag(f"v{self.version}", force=True)
+            remote.push(refspec=tag_ref)
+
+        return remote.push()[0].summary.split("..")[1].rstrip()
+
+    def get_changed_files(self):
+        """ Get a list of changed files compared to remote. """
+        repo = Repo(str(self.path))
+        return re.findall("diff --git a/(.*) " + "b/", repo.git.diff())
 
     def bump_version(self):
         """ Bump micro version in metadata.json. """
-        parts = self.version.split(".")
-        parts[-1] = str(int(parts[-1]) + 1)
-        self.version = ".".join(parts)
+        self.version = self.version.bump()
 
+    def pip_install(self):
+        """ Install this repository with pip, WITHOUT -e flag.
+            Subprocess messed up -e flag compared to doing it in terminal, so use the normal one."""
+        # subprocess.call(f"python -m pip install {self.path}")
+        subprocess.check_call([sys.executable, "-m", "pip", "install", str(self.path)])
 
+    def unittest(self):
+        """ Run unittests for this repository. """
+        # subprocess.call(f"python -m unittest discover {self.get_test_path()}")
+        subprocess.check_call([sys.executable, "-m", "unittest", "discover", str(self.get_test_path())])
+
+    def create_sdist(self):
+        """ Create source distribution. """
+        with self.path.as_working_dir():
+            subprocess.check_call([sys.executable, "setup.py", "sdist", "bdist_wheel"])
+
+    def upload(self):
+        """ Upload local repo to PyPI. """
+        self.create_sdist()
+        with self.path.as_working_dir():
+            subprocess.check_call([sys.executable, "-m", "twine", "upload", "dist/*"])
+
+for key in LocalRepo.metadata_keys:
+    setattr(LocalRepo, key, property(
+        lambda self, key=key: getattr(self, f"_{key}", ...),
+        lambda self, value, key=key: LocalRepo.metadata_setter(self, key, value),
+    ))
 
 
 
